# proj62-Linux锁优先级传递

时间：2024.05.29

选题：[proj-62 Linux锁优先级传递](https://github.com/oscomp/proj62-linux-lock-priority-inheritance)

队伍：我爱说实话

队伍成员：曲晨阳，卫策，王淑华

## 1. 比赛准备和调研

### 1.1 题目描述：
优先级翻转是指当一个高优先级任务通过锁等机制访问共享资源时, 该锁如果已被一低优先级任务占有, 将可能造成高优先级任务被许多具有较低优先级任务阻塞. 优先级翻转问题是影响linux实时性的障碍之一, 甚至引起很多稳定性问题.
该项目的目标:
- 实现rwlock、rwsem、mutex的优先级继承支持
- 通过内核宏开关优先级继承相关特性

### 1.2 什么是优先级翻转
优先级反转是指低优先级任务先于高优先级任务执行的现象，具体来说就是高优先级任务被低优先级任务阻塞，导致高优先级任务迟迟得不到调度。但其他中等优先级的任务却能抢到CPU资源。从运行情况上来看，好像是中优先级的任务比高优先级任务具有更高的优先权。

举一个具体的例子，假定一个进程中有三个线程Thread1(高）、Thread2（中）和Thread3（低），考虑下图的执行情况。

![优先级反转](./img/优先级反转.png)

其每个时刻的运行过程如下：
- T0时刻，Thread3运行，并获得同步资源SYNCH1；
- T1时刻，Thread2开始运行，由于优先级高于Thread3，Thread3被抢占（未释放同步资源SYNCH1），Thread2被调度执行；
- T2时刻，Thread1抢占Thread2；
- T3时刻，Thread1需要同步资源SYNCH1，但SYNCH1被更低优先级的Thread3所拥有，Thread1被挂起等待该资源优先级的Thread3所拥有，Thread1被挂起等待该资源

此时线程Thread2和Thread3都处于可运行状态，Thread2的优先级大于Thread3的优先级，Thread2被调度执行。最终的结果是高优先级的Thread1迟迟无法得到调度，而中优先级的Thread2却能抢到CPU资源。

### 1.3 优先级翻转的解决方法
解决方法有两种：优先级天花板（priority ceiling，PC）和优先级继承（priority inheritance，PI）。

优先级天花板（priority ceiling）是指将申请某资源的任务的优先级提升到可能访问该资源的所有任务中最高优先级任务的优先级。每个临界资源R赋予一个优先级天花板，取所有可能访问临界资源R的所有进程中优先级最高的作为R的优先级天花板。当一个进程获取到临界资源，就将该进程的优先级提升到优先级天花板，这样，该进程不会被其他可能使用该临界资源的进程抢占，从而得到更快执行，更快地释放临界资源。释放临界资源后，恢复优先级。

优先级继承（priority inheritance）是指当高优先级进程（P1）请求一个已经被被低优先级（P3）占有的临界资源时，将低优先级进程（P3）的优先级临时提升到与高优先级进程一样的级别，使得低优先级进程能更快地运行，从而更快地释放临界资源。低优先级进程离开临界区后，其优先级恢复至原本的值。回到上述例子，拥有优先级继承的线程调度如下图所示。

![优先级继承](./img/优先级继承.png)

到了T3时刻，Thread1需要Thread3占用的同步资源SYNCH1，操作系统检测到这种情况后，就把Thread3的优先级提高到Thread1的优先级。此时处于可运行状态的线程Thread2和Thread3中，Thread3的优先级大于Thread2的优先级，Thread3被调度执行。

Thread3执行到T4时刻，释放了同步资源SYNCH1，操作系统恢复了Thread3的优先级，Thread1获得了同步资源SYNCH1，重新进入可执行队列。处于可运行状态的线程Thread1和Thread2中，Thread1的优先级大于Thread2的优先级，所以Thread1被调度执行。

## 2. 算法设计原理
本项目旨在实现rwlock、rwsem、mutex的优先级继承（PI）支持。如上所述，rwlock和rwsem中的读任务共享临界区，因此本项目直接继承原有的read_lock和read_sem设计，主要针对write_lock、write_sem和mutex添加优先级继承支持，其算法的设计思路类似，以下统一介绍其原理。

### 2.1 优先级继承链（PI chain）
优先级继承是指锁的低优先级所有者继承高优先级等待者的优先级，直到锁被释放。而在实际问题中，往往不像上述的三个任务的例子那么简单，举一个复杂一点的例子，假设有五个进程：P1、P2、P3、P4、P5，有四个锁：L1、L2、L3、L4，其运行情况如下：
- P1拥有锁L1
- P2拥有锁L2，等待L1释放行；
- P3拥有锁L3，等待L2释放
- P4拥有锁L4，等待L3释放
- P4拥有锁L4，等待L3释放
- P5等待L4释放

可见多个锁和多个进程相互阻塞，可以抽象为一个阻塞链：

![PIchain1](./img/PIchain1.png)

在这条阻塞链中，假如有优先级继承支持，那么P1的优先级应来着这条链上的最大优先级。

而实际问题中，往往存在多个进程被一个锁阻塞，则阻塞链实际上为树形结构，假如在以上的例子中添加一个新进程P6和一个新锁L5，P6请求获取L5，而L5被P2持有，则P6被P2阻塞无法运行，阻塞链则更新为：

![PIchain2](./img/PIchain2.png)

若此锁拥有优先级继承功能，则要求链的最右侧，也就是链的顶端的优先级始终大于等于左侧的优先级，这就是优先级继承链（PI chain）。

在使用有优先级继承的锁时，始终维护这一个PI chain，实现优先级的继承与更新。

### 2.2 等待者树
每个锁使用一个红黑树（rbtree）按优先级存储这些等待者。该树受到位于互斥体结构中的自旋锁的保护。这个自旋锁为wait_lock。

### 2.3 优先级继承树
每个进程使用一个红黑树保存被该进程的锁阻塞的最高优先级等待者，称为优先级继承树（PI rbtree）,该树只包含顶部等待者，也就是等待该锁的最高优先级进程，如果进程继承了优先级，那么位于该树顶部的进程的优先级将始终是该进程的优先级。

## 3. 运行结果演示
本项目基于Linux Kernel 4.19.311版本实现，以Linux内核模块的形式将添加了优先级继承的pi_rwlock、pi_rwsem、pi_mutex添加进内核中，使用EXPORT_SYMBOL_GPL()方法将其作为符号导出提供给外部模块使用。通过修改Kconfig文件将其加入make menu，在内核编译的时候可以选择是否编译这三个模块，选中的时候如下图所示：

![menuconfig](./img/menuconfig.png)

本项目的测试程序使用第一章的例子进行设计，按顺序创建低、高、中三个优先级的线程，低优先级线程先获得锁，然后高优先级线程请求锁被阻塞，中优先级线程尝试抢占低优先级线程的运行。若优先级顺利继承，低优先级线程将继承来自高优先级线程的优先级，不会被中优先级线程抢占运行。

实验结果如下图所示，可见在低优先级线程释放锁后，高优先级线程获得锁继续运行，其中没有被中优先级线程抢占，说明优先级被顺利继承。

![pilocktest](./img/pilocktest.png)

下图是使用原本的没有优先级继承的锁，可见中等优先级线程抢占了低优先级线程，导致高优先级线程在中优先级线程之后运行。

![rawlock_test](./img/rawlock_test.png)