
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/kdev_t.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include<linux/slab.h>                 //kmalloc()
#include<linux/uaccess.h>              //copy_to/from_user()
#include <linux/kthread.h>             //kernel threads
#include <linux/sched.h>               //task_struct 
#include <linux/delay.h>
// #include <linux/mutex.h>
#include <linux/pimutex.h>
#include <linux/err.h>

struct pi_mutex etx_pimutex;

unsigned long etx_global_variable = 0;
 
dev_t dev = 0;
static struct class *dev_class;
static struct cdev etx_cdev;
 
static int __init etx_driver_init(void);
static void __exit etx_driver_exit(void);
 
static struct task_struct *etx_thread_low;
static struct task_struct *etx_thread_high; 
static struct task_struct *etx_thread_mid; 

/*************** Driver functions **********************/
static int etx_open(struct inode *inode, struct file *file);
static int etx_release(struct inode *inode, struct file *file);
static ssize_t etx_read(struct file *filp, 
                char __user *buf, size_t len,loff_t * off);
static ssize_t etx_write(struct file *filp, 
                const char *buf, size_t len, loff_t * off);
 /******************************************************/
 
int thread_function_low(void *pv);
int thread_function_high(void *pv);
int thread_function_mid(void *pv);

/*
** Thread function 1
*/
int thread_function_low(void *pv)
{
    
    while(!kthread_should_stop()) {
        pi_mutex_lock(&etx_pimutex);
        pr_info("thread_low start running, get mutex!\n");
        mdelay(10000);
        etx_global_variable++;
        pr_info("In EmbeTronicX Thread Function1 %lu\n", etx_global_variable);
        pi_mutex_unlock(&etx_pimutex);
        pr_info("thread_low stop running, free mutex!\n");
    }
    return 0;
}

/*
** Thread function 2
*/
int thread_function_high(void *pv)
{
    while(!kthread_should_stop()) {
        pr_info("thread_high waiting mutex...\n");
        pi_mutex_lock(&etx_pimutex);
        pr_info("thread_high start running, get mutex!\n");
        etx_global_variable++;
        pr_info("In EmbeTronicX Thread Function_high %lu\n", etx_global_variable);
        pi_mutex_unlock(&etx_pimutex);
        pr_info("thread_high stop running, free mutex!\n");
        kthread_stop(etx_thread_high);
    }
    return 0;
}

int thread_function_mid(void *pv)
{
    
    while(!kthread_should_stop()) {
        pr_info("Thread_mid is running!\n");
        mdelay(5000);
        pr_info("Thread_mid running over!\n");
        kthread_stop(etx_thread_mid);
    }
    return 0;
}

//File operation structure
static struct file_operations fops =
{
        .owner          = THIS_MODULE,
        .read           = etx_read,
        .write          = etx_write,
        .open           = etx_open,
        .release        = etx_release,
};

/*
** This function will be called when we open the Device file
*/ 
static int etx_open(struct inode *inode, struct file *file)
{
        pr_info("Device File Opened...!!!\n");
        return 0;
}

/*
** This function will be called when we close the Device file
*/
static int etx_release(struct inode *inode, struct file *file)
{
        pr_info("Device File Closed...!!!\n");
        return 0;
}

/*
** This function will be called when we read the Device file
*/ 
static ssize_t etx_read(struct file *filp, 
                char __user *buf, size_t len, loff_t *off)
{
        pr_info("Read function\n");
 
        return 0;
}

/*
** This function will be called when we write the Device file
*/
static ssize_t etx_write(struct file *filp, 
                const char __user *buf, size_t len, loff_t *off)
{
        pr_info("Write Function\n");
        return len;
}

/*
** Module Init function
*/
static int __init etx_driver_init(void)
{
        /*Allocating Major number*/
        if((alloc_chrdev_region(&dev, 0, 1, "etx_Dev")) <0){
                pr_info("Cannot allocate major number\n");
                return -1;
        }
        pr_info("Major = %d Minor = %d \n",MAJOR(dev), MINOR(dev));
 
        /*Creating cdev structure*/
        cdev_init(&etx_cdev,&fops);
 
        /*Adding character device to the system*/
        if((cdev_add(&etx_cdev,dev,1)) < 0){
            pr_info("Cannot add the device to the system\n");
            goto r_class;
        }
 
        /*Creating struct class*/
        if(IS_ERR(dev_class = class_create(THIS_MODULE,"etx_class"))){
            pr_info("Cannot create the struct class\n");
            goto r_class;
        }
 
        /*Creating device*/
        if(IS_ERR(device_create(dev_class,NULL,dev,NULL,"etx_device"))){
            pr_info("Cannot create the Device \n");
            goto r_device;
        }
 
        pi_mutex_init(&etx_pimutex);

        /* Creating Thread 1 */
        // etx_thread_low = kthread_run(thread_function_low,NULL,"eTx Thread1");
        etx_thread_low = kthread_create(thread_function_low, NULL, "eTx Thread1");
        if(etx_thread_low) {
            kthread_bind(etx_thread_low, 0);
            set_user_nice(etx_thread_low, 15);
            mdelay(2000);
            pr_err("Kthread_low Created Successfully...\n");
            wake_up_process(etx_thread_low);
        } else {
            pr_err("Cannot create kthread_low\n");
             goto r_device;
        }
 
         /* Creating Thread 2 */
        // etx_thread_high = kthread_run(thread_function_high,NULL,"eTx Thread2");
        etx_thread_high = kthread_create(thread_function_high, NULL, "eTx Thread2");
        if(etx_thread_high) {
		    kthread_bind(etx_thread_high, 1);
            set_user_nice(etx_thread_high, 5);
            mdelay(2000);
            pr_err("Kthread_high Created Successfully...\n");
            wake_up_process(etx_thread_high);
        } else {
            pr_err("Cannot create kthread_high\n");
             goto r_device;
        }

        etx_thread_mid = kthread_create(thread_function_mid, NULL, "eTx Thread3");
        if(etx_thread_mid) {
            kthread_bind(etx_thread_mid, 0);
            set_user_nice(etx_thread_mid, 10);
            mdelay(2000);
            pr_err("Kthread_mid Created Successfully...\n");
            wake_up_process(etx_thread_mid);
        } else {
            pr_err("Cannot create kthread_mid\n");
             goto r_device;
        }
        
        pr_info("Device Driver Insert...Done!!!\n");
        return 0;
 
 
r_device:
        class_destroy(dev_class);
r_class:
        unregister_chrdev_region(dev,1);
        cdev_del(&etx_cdev);
        return -1;
}

/*
** Module exit function
*/ 
static void __exit etx_driver_exit(void)
{
        kthread_stop(etx_thread_low);
        kthread_stop(etx_thread_high);
        kthread_stop(etx_thread_mid);
        device_destroy(dev_class,dev);
        class_destroy(dev_class);
        cdev_del(&etx_cdev);
        unregister_chrdev_region(dev, 1);
        pr_info("Device Driver Remove...Done!!\n");
}
 
module_init(etx_driver_init);
module_exit(etx_driver_exit);
 
MODULE_LICENSE("GPL");
MODULE_VERSION("1.17");
